/*
 * Accurate binary image selection from inaccurate user input
 *
 * Authors: Kartic Subr, Sylvain Paris, Cyril Soler, Jan Kautz
 *
 *
 * Copyright (2013):Kartic Subr. 
 * All rights reserved.
 *
 * For details, see the paper:
 * Accurate binary image selection from inaccurate user input 
 * Computer Graphics Forum 32, 2 (2013)
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for educational, research, and non-commercial
 * purposes, without fee and without a signed licensing agreement, is
 * hereby granted, provided that the above copyright notice and this
 * paragraph appear in all copies modifications, and distributions.
 *
 * Any commercial use or any redistribution of this software
 * requires a license.
 *
 * For further details, contact Kartic Subr (kartic@gmail.com)
 */

#include "densecrf.h"
#include <cstdio>
#include <cmath>
#include <Magick++.h> 
#include <DataMat.h>
#include <iostream>
#include <fstream>
#include <FactoredImage.h>
#include <sstream>

using std::stringstream; 
using namespace std ;
extern string IMPATH  ;
using namespace Magick; 

// Simple classifier that assigns label probabilities (f,b,v), for each pixel
// (foreground, background, void) 
// if the pixel is marked as fg, then (0.5, 0.25, 0.25)
// if the pixel is marked as bg, then (0.25, 0.5, 0.25)
// if the pixel is not marked, then (0.33, 0.33, 0.33)
float * FBVunary1( const vector<vector<classifFBV> >& Labels, int W, int H, int M )
{
    if (!Labels.size()) return 0 ;
    
    const int w(Labels[0].size()), h(Labels.size()) ;
    const float GT_PROB=0.5 ;
    const float u_energy = -log( 1.0 / M );
    const float n_energy = -log( (1.0 - GT_PROB) / (M-1) );
    const float p_energy = -log( GT_PROB );
    float * res = new float[W*H*M];
    for (int x(0); x<w; x++)
    {
	for(int y(0); y<h; y++)
	{
	    int i(0); 
	    switch(Labels[y][x])
	    {
		case FG:
		    i=1;
		    break;
		    
		case BG:
		    i=2;
		    break; 
	    }
	    
	    int k = y*w + x ;
	    // Set the energy
	    float * r = res + k*M;
	    if (i){
		    for( int j=0; j<M; j++ )
			    r[j] = n_energy;
		    r[i] = p_energy;
	    }
	    else
	    {
		    for( int j=0; j<M; j++ )
			    r[j] = u_energy;
	    }
	}
    }
    return res;
}


void ReadFGBGMat(const string& f, const string& b, DataMat& F, DataMat& B, int htarget, int wtarget) 
{
	ifstream iF(f.c_str()), iB(b.c_str()) ;
	if (!iF || !iB) 
	{
	    cerr << "Could not open fg/bg scribbles file. Check " << f << " and " << b << endl ;
	    exit (-1);
	}
	
	int w(0), h(0), dmy(0); 
	iF >> h ; 
	iB >> dmy ;
	if (h!=dmy) {cerr << "FG and BG scrib heights are incompatible: " << h << " " << dmy << endl ; exit(1) ;}
	iF >> w ; 	
	iB >> dmy ;
	if (w!=dmy) {cerr << "FG and BG scrib widths are incompatible: " << w << " " << dmy <<  endl ; exit(1) ;}
	
	F.resize(h,w) ;
	B.resize(h,w) ;
	
	while (iF)
	{
	    int r(0), c(0) ;
	    iF >> r ; 	    
	    if (!iF) {break; }
	    iF >> c ;    
	    F(r-1,c-1) = 1; 	    
	}
	
	while (iB)
	{
	    int r(0), c(0) ;
	    iB >> r ; 	    
	    if (!iB) {break;}
	    iB >> c ;
	    B(r-1,c-1) = 1; 
	}
	
	// Resize F and B to be htarget x wtarget
	// ... 
}

void PFMtoUCharArr(const PFMImage& I, unsigned char*& im) 
{
    const int h(I.height), w(I.width) ;
    int maxval(0) ;
    for (int i(0); i<h; i++) 
    	for (int j(0); j<w; j++) 
	{
	    PFMPixel p = I(i,j) ;
	    maxval = max(p.r, max(p.g, p.b)) ;
	}
	    
    if (fabs(maxval)<1e-2) maxval=1; 
    im = new unsigned char [w*h*3] ;
    for (int i(0); i<h; i++) 
    {
	for (int j(0); j<w; j++) 
	{
	    PFMPixel p = I(i,j) ;
	    im[((i*w)+j)*3] = static_cast<unsigned char>(floor(p.r/maxval*255)) ;
	    im[((i*w)+j)*3+1] = static_cast<unsigned char>(floor(p.g/maxval*255)) ;
	    im[((i*w)+j)*3+2] = static_cast<unsigned char>(floor(p.b/maxval*255)) ;
// 	    cout << (int)im[i*j*3] << " " << flush ;
	}
    }
}


///////////////////////////////////////// 
void CallCRFVersion(int argc, char* argv[], const PFMImage& I, FactoredImage* pFI, const string& fg, const string& bg, const string& o, int ds, double sigma, double wt, bool distComp)  
{
        
    const int M(3) ;
    const double distThresh(0.0) ;
    const int h(pFI->GetHeight()), w(pFI->GetWidth()) ;
    DataMat BGMat, FGMat ;
    ReadFGBGMat(fg, bg, FGMat, BGMat, h,w) ;
    
    unsigned char *im, *anno ;    
    vector<vector<classifFBV> > Labels; 
    
    cout << "Building Labels..." << endl ;    
    
	Labels.clear() ;
	vector<classifFBV> dummy(w, VD) ;
	for (int i(0); i<h; i++)
	    Labels.push_back(dummy) ;
	
	for (int i(0); i<h; i++)
	{
	    for(int j(0); j<w; j++)
	    {			  
		if (BGMat(i,j)>0.5) Labels[i][j]=BG; 
		if (FGMat(i,j)>0.5) Labels[i][j]=FG ;
	    }
	}
    
        
    cout << "Constructing unary potential ..." << endl ;
    float* unary = FBVunary1( Labels, w, h, M );
        
    DenseCRF2D crf(w, h, M);    
    crf.setUnaryEnergy( unary);    
    
    PFMtoUCharArr(pFI->GetOriginalImage(), im) ;
//     crf.addPairwiseBilateral( 60, 60, 20, 20, 20, im, 10 );

    cout << "Adding factored image..." << endl ;    
    crf.addFactoredImage( sigma, pFI, im, wt );
    
    cout << "Constructing map ..." << endl ;
    short * map = new short[w*h];
    crf.map(10, map);
    cout << "Setting influence maps with refined regions ..." << endl ;
    
    unsigned char* px = new unsigned char [w*h*3] ;
    unsigned char* px2 = new unsigned char [w*h*3] ;
    for (int i(0); i<h; i++)
    {
	for(int j(0); j<w; j++)
	{
	    unsigned char rgbvals[3] = {0, 0, 0};
	    int idx=3*(i*w+j);	    
	     if (map[i*w+j] == 1) rgbvals[2] = 255 ; 	    
	     if (map[i*w+j] == 2) rgbvals[0] = 255 ;		 	    
	    px[idx] = rgbvals[0] ;     px[idx+1] = rgbvals[1] ;     px[idx+2] = rgbvals[2] ;
	    
	    unsigned char rgbvals2[3] = {0, 0, 0};
	    if (Labels[i][j]==FG)  rgbvals2[2] = 255 ; 
	    if (Labels[i][j]==BG)  rgbvals2[0] = 255 ;     
	    px2[idx] = rgbvals2[0] ;     px2[idx+1] = rgbvals2[1] ;     px2[idx+2] = rgbvals2[2] ;

	}
    }    
    ostringstream oss ;
    oss << o << ".sig." << sigma << ".wt." << wt << ".FGBG.png" ;  
    cout << "Writing output to " << oss.str() ;
    Image ImgOut(w,h,"RGB", CharPixel, px) ;
    ImgOut.write(oss.str()) ;
    
    Image ImgOut2(w,h,"RGB", CharPixel, px2) ;
    ImgOut2.write(oss.str()+".inp.png") ;
    
    delete [] px2 ;
    delete [] px ;
    delete [] map ;
    delete [] unary ;
    delete[] anno;
    delete[] im;

}
