/*
 * Accurate binary image selection from inaccurate user input
 *
 * Authors: Kartic Subr, Sylvain Paris, Cyril Soler, Jan Kautz
 *
 *
 * Copyright (2013):Kartic Subr. 
 * All rights reserved.
 *
 * For details, see the paper:
 * Accurate binary image selection from inaccurate user input
 * Computer Graphics Forum 32, 2 (2013)
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for educational, research, and non-commercial
 * purposes, without fee and without a signed licensing agreement, is
 * hereby granted, provided that the above copyright notice and this
 * paragraph appear in all copies modifications, and distributions.
 *
 * Any commercial use or any redistribution of this software
 * requires a license.
 *
 * For further details, contact Kartic Subr (kartic@gmail.com)
 */

#include <DataMat.h>
#include <iostream>
#include <cstdlib>
#include <fstream>
#include <cassert>
#include <iostream>
#include <cmath>
#include <string>
#include <vector>
#include <iterator>
#include <sstream>
#include <fstream>
#include <PFMImage.h>

typedef double Real ;

namespace
{
	inline bool CloseToZero(double d)
	{
		return fabs(d) < 1e-3 ;
	}
	
	template<typename Real>
			inline Real clamp(const Real min_value, const Real max_value, const Real x)
	{ return std::max(std::min(x,max_value),min_value); }
        
	inline double bilinear_interpolation(const DataMat& array, const Real x, const Real y)
	{
		typedef unsigned int size_type;
		typedef double real_type;
                
		const size_type x_size = array.x_size();
		const size_type y_size = array.y_size();
                
		const size_type x_index  =
				clamp<size_type>(0,x_size-1,static_cast<size_type>(x));
		const size_type xx_index =
				clamp<size_type>(0,x_size-1,x_index+1);
                
		const size_type y_index  =
				clamp<size_type>(0,y_size-1,static_cast<size_type>(y));
		const size_type yy_index =
				clamp<size_type>(0,y_size-1,y_index+1);
                
		const real_type x_alpha = x - x_index;
		const real_type y_alpha = y - y_index;
                
		return
				(1.0-x_alpha) * (1.0-y_alpha) * array(y_index, x_index) +
				x_alpha       * (1.0-y_alpha) * array(y_index,xx_index) +
				(1.0-x_alpha) * y_alpha       * array(yy_index, xx_index) +
				x_alpha       * y_alpha       * array(yy_index,xx_index);                        
	}

}

// assume datamat values are in [0,1]
// for nBins in MaxVal-MinVal, the bin centers are dw = (MaxVal-MinVal)/nBins units apart
// the bin centres are at (2p+1)dw/2 where p = 0, 1, 2 ... nBins-1
// For a given value v, its bin # is k=floor(v/dw)
// Distance to its center is d = v-(2k+1)dw/2
// Contribution of v = 1-d/dw for bin # k and d/dw for bin # k+/-1 

void DataMat::Histogram(vector<double>& hist, vector<double>& CumulHist, int nBins, bool normalize) const
{
	if (nBins<=0) return; 
// 	hist.resize(nBins) ;
// 	CumulHist.resize(nBins) ;
	double MinVal(_V[0][0]), MaxVal(_V[0][0]) ;
	int nR(NRows()), nC(NCols()) ;
	for (int r(0); r<nR; r++)
	{                
		for (int c(0); c<nC; c++)
		{ MinVal=min(MinVal, _V[r][c]); MaxVal=max(MaxVal, _V[r][c]);}
	}
	MinVal = 0 ;
	MaxVal = 1; 	
	double dw((MaxVal-MinVal)/nBins) ;
	
// 	With Linear inteprolation
	for (int r(0); r<nR; r++)
	{                
		for (int c(0); c<nC; c++)
		{
			int k = max(0, min((int)floor(_V[r][c]/dw), nBins-1)) ;
			if (k<0 || k>=nBins) cout << k << " " << _V[r][c] << endl ;
			assert (k>=0 && k<nBins) ;
			
			double d = _V[r][c]-(2*k+1)*dw/2 ;
			int l = (d>0) ? k+1: k-1;
			l = max(0, min(nBins-1, l)) ;
			hist[l] += fabs(d/dw);
			hist[k] += 1-fabs(d/dw) ;
// 			if (d < 0)
// 			{
// 				cout << k << " " << d << " " << dw << " " << _V[r][c] << " " << (2*k+1)*dw/2 << " " << _V[r][c]-(2*k+1)*dw/2 << " " << fabs(d/dw) << " " << 1-fabs(d/dw) << endl ;
// 			} 
		}
	}
	
// 	Without Linear interpolation
// 	for (int r(0); r<nR; r++)
// 	{                
// 		for (int c(0); c<nC; c++)
// 		{
// 			int loc = min((int)floor(_V[r][c]*dw), nBins-1) ;
// 			if (loc<0 || loc>=nBins) cout << loc << " " << _V[r][c] << endl ;
// 			assert (loc>=0 && loc<nBins) ;
// 			hist[loc]+=1.0; 
// 		}
// 	}		
	int KSize(5) ;
	double Var(1) ;
	vector<double> oldhist(hist) ;
	for (int i(0); i<nBins; i++)
	{
		double blur (0), w(0)  ;
// 		cout << " i = " << i << endl ;
		for (int j(0); j<KSize; j++)
		{
			int k = i-KSize/2+j ;
			if (k<0 || k>=nBins) continue ;
// 			cout << " \t\t j " << j << " " << k << endl ; 
			const double wt = exp(-Var*pow((i-j)/(1.0*KSize),2.0)) ;
			blur += oldhist[k] * wt ; w += wt ;
			
// 			blur += oldhist[k] ;
// 			w ++ ;
		}
		if (w>1e-9) 
			hist[i] = blur/w ;
	}
// 		cin.ignore() ;
	double SumHist(0) ;
	for (int i(0); i<nBins; i++)
	{
		if (normalize) hist[i] /= (nR*nC) ;
		SumHist += hist[i] ;
		CumulHist[i] = SumHist ;
	}
	
}

                
DataMat::DataMat(int nRows, int nCols) 
{
	vector <double> V(nCols, 0.0) ;
	for (int i(0); i<nRows; i++) _V.push_back(V) ;
}

DataMat::DataMat(const DataMat& mat,int nRows, int nCols) 
{
	vector <double> V(nCols, 0.0) ;
	for (int i(0); i<nRows; i++) 
	{
		_V.push_back(V) ;
	
		for(int j=0;j<nCols;++j)
			operator()(i,j) = mat.Interp( i/(double)nRows * mat.NRows(), j/(double)nCols * mat.NCols() ) ;
	}
}

double DataMat::Interp(double i,double j) const
{
	int I = (int)floor(i) ;
	int J = (int)floor(j) ;
	double di(i-I) ;
	double dj(j-J) ;

	int Ip1 = min(I+1,NRows()-1) ;
	int Jp1 = min(J+1,NCols()-1) ;

	return (1.0 - di) * ( (1.0 - dj) * operator()(I  ,J) + dj * operator()(I  ,Jp1) )
			+ di  * ( (1.0 - dj) * operator()(Ip1,J) + dj * operator()(Ip1,Jp1) ) ;
}

DataMat::DataMat(const DataMat& DM) 
{               
	int nR(DM.NRows()) ;                
	for (int i(0); i<nR; i++) _V.push_back(DM._V[i]) ;
}

void DataMat::resize(const DataMat& D) 
{
	int nR(D.NRows()) ;                
	_V.resize(D._V.size()) ;
	for (int i(0); i<nR; i++) _V[i].resize(D._V[i].size()) ;
}
                                
const DataMat&  DataMat::operator += (const DataMat& D) 
{
	int minNR(min(NRows(), D.NRows())), minNC(min(NCols(), D.NCols())) ;
	for (int i(0); i<minNR; i++)
		for(int j(0); j<minNC; j++)
			_V[i][j] += D._V[i][j] ;

	return *this ;
}

const DataMat&  DataMat::operator *= (const DataMat& D) 
{
	int minNR(min(NRows(), D.NRows())), minNC(min(NCols(), D.NCols())) ;
	for (int i(0); i<minNR; i++)
		for(int j(0); j<minNC; j++)
			_V[i][j] *= D._V[i][j] ;

	return *this ;
}

const DataMat&  DataMat::Assign (const DataMat& D, double d) 
{
	int minNR(min(NRows(), D.NRows())), minNC(min(NCols(), D.NCols())) ;        
	for (int i(0); i<minNR; i++)
		for(int j(0); j<minNC; j++)
			_V[i][j] = d*D._V[i][j] ;

	return *this ;
}

const DataMat&  DataMat::operator = (const DataMat& D) 
{
    _V = D._V ;
	return *this ;
}

const DataMat&  DataMat::operator -= (const DataMat& D) 
{
	int minNR(min(NRows(), D.NRows())), minNC(min(NCols(), D.NCols())) ;
	for (int i(0); i<minNR; i++)
		for(int j(0); j<minNC; j++)
			_V[i][j] -= D._V[i][j] ;

	return *this ;
}

const DataMat&  DataMat::operator -= (double d) 
{
    	for (int i(0); i<NRows(); i++)
		for(int j(0); j<NCols(); j++)
			_V[i][j] -= d ;
    return *this; 
}

const DataMat&  DataMat::operator *= (double d) 
{
    	for (int i(0); i<NRows(); i++)
		for(int j(0); j<NCols(); j++)
			_V[i][j] *= d ;
    return *this ;
}

double DataMat::GetDR() const 
{        
	int NR(NRows()), NC(NCols()) ;
	if (!NR) return 0 ;
	double Max(_V[0][0]), Min(_V[0][0]) ;
	for (int i(0); i<NR; i++)
		for(int j(0); j<NC; j++)
	{
		Max = max(_V[i][j], Max) ;
		Min = min(_V[i][j], Min) ;
	}
	return Max-Min ;
}

void DataMat::Open(const string& Fname) 
{
	int nR(0), nC(0); 
	ifstream ifsbin(Fname.c_str(), ios::binary) ;
        if (!ifsbin) cerr << "Cannot open file " << Fname << endl ;
	ifsbin.read((char*)&nR, sizeof(int)) ;
	ifsbin.read((char*)&nC, sizeof(int)) ;
// 	cout << nR  << " " << nC << endl ;
	resize(nR, nC); 
	double *d = new double[nR*nC] ;
	ifsbin.read((char*)d, sizeof(double)*nR*nC) ;
	for (int r(0), i(0); r<nR; r++) for(int c(0); c<nC; c++) _V[r][c]=d[i++] ;
	ifsbin.close() ;
	delete[] d ;
}

void DataMat::BinaryWrite(const string& fname, bool Verbose) const
{
	if (_V.size()==0) return; 
	int nR(NRows()), nC(NCols()) ;
	if (Verbose) cout << "Writing " << fname << " of size " << nR << " x " << nC << endl ;
	ofstream ofsbin(fname.c_str(), ios::binary) ;
	ofsbin.write((char*)&nR, sizeof(int)) ;
	ofsbin.write((char*)&nC, sizeof(int)) ;
	double *arr = new double[nR*nC] ;
	for (int r(0), i(0); r<nR; r++) for(int c(0); c<nC; c++) arr[i++]=_V[r][c]; 
	ofsbin.write((char*)arr, sizeof(double)*nR*nC) ;
	ofsbin.close(); 

	delete[] arr ;
}

void DataMat::NormalizedGrad(DataMat& GradNorm) const
{
        int NR(NRows()), NC(NCols()) ;
        double eps(.01) ;
        for (int r(2); r<NR-2; r++)
        {                
                for (int c(2); c<NC-2; c++)
                {        
                        double dx =  _V[r][c+1] - _V[r][c-1] ;                        
                        double dy =  _V[r+1][c] - _V[r-1][c] ;
                                                
                        double dx5 =  - 0.0828*_V[r][c+2] + 0.66561*_V[r][c+1]  - 0.66561*_V[r][c-1] + 0.0828*_V[r][c-2] ; 
                        double dy5 =  - 0.0828*_V[r+2][c] + 0.66561*_V[r+1][c]  - 0.66561*_V[r-1][c] + 0.0828*_V[r-2][c] ;        
                        
//                         dx = dx5; dy = dy5 ;
                        
                        double n = sqrt(dx*dx + dy*dy) ;
                        if (n<1e-10) n = 1 ;
                        GradNorm(r,c) = n ;
                }
        }
}

double DataMat::MaxVal() const 
{
    int NR(NRows()), NC(NCols()) ;
    if (!(NR+NC)) return 0 ;
    double maxval(0.0) ;
    maxval = _V[0][0] ;
    for (int r(0); r<NR; r++)
	{                
		for (int c(0); c<NC; c++)
		{
		    maxval = max(maxval,_V[r][c]) ;
		}
	}
	return maxval ;
}


