/*
 * Accurate binary image selection from inaccurate user input
 *
 * Authors: Kartic Subr, Sylvain Paris, Cyril Soler, Jan Kautz
 *
 * Copyright (2013):Kartic Subr. 
 * All rights reserved. 
 *
 * For details, see the paper:
 * Accurate binary image selection from inaccurate user input
 * Computer Graphics Forum 32, 2 (2013)
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for educational, research, and non-commercial
 * purposes, without fee and without a signed licensing agreement, is
 * hereby granted, provided that the above copyright notice and this
 * paragraph appear in all copies modifications, and distributions.
 *
 * Any commercial use or any redistribution of this software
 * requires a license.
 *
 * For further details, contact Kartic Subr (kartic@gmail.com)
 */

#include <DistHelper.h>
#include <DataMat.h>
#include <PFMImage.h>
#include <FactoredImage.h>
#include <Hist.h>
#include <omp.h>
#include <iomanip>

DistMetric int2DistMetric[15] = {EUC, LAP, BIL, FAB, PAT, SSIM, L1, CHISQ,EMD, DMAP, BIH, CORREL } ; 
 

template<>
EnumParser<DistMetric>::EnumParser()
{
    enumMap["EUC"] = EUC;
    enumMap["LAP"] = LAP;
    enumMap["BIL"] = BIL;
    enumMap["FAB"] = FAB;
    enumMap["PAT"] = PAT;
    enumMap["SSIM"] = SSIM;
    enumMap["L1"] = L1;
    enumMap["CHISQ"] = CHISQ;
    enumMap["EMD"] = EMD;
    enumMap["DMAP"] = DMAP;
    enumMap["BIH"] = BIH;
    enumMap["CORREL"] = CORREL;
}
 
DistHelper::DistHelper(const PFMImage& I, DistMetric distTYPE, double sr, double ss, bool LUVSpace, int PatchSize) : _dtype (distTYPE), _h(I.height), _w(I.width),   _I(I), _sigsqr(sr*sr), _sigsqs(ss*ss), ExistFIs(false)
{
    if (_h*_w==0) return ; 
    const double eps = 1e-6 ;
    if (distTYPE==SSIM) LUVSpace==true; 
    
    // Luv instead of RGB
    if (LUVSpace)
    {
	double MaxL (0) ;
	double MinL (0) ;
	 for(int i(0); i<_h; i++)
	    for(int j(0); j<_w; j++)
	    {
		PFMPixel p = _I(i,j) ;
		_I(i,j) = RGB2LUV(p) ; 
		MaxL = max(MaxL, (double)_I(i,j).r) ;
		MinL = min(MinL, (double)_I(i,j).r) ;
	    }
	  
	 double DivLdr(1.0/(MaxL-MinL)) ;
	 if (fabs(MaxL-MinL) <1e-8) DivLdr = 1.0 ;
	 
	 for(int i(0); i<_h; i++)
	    for(int j(0); j<_w; j++)
		    _I(i,j) = DivLdr*(_I(i,j)-PFMPixel(MinL,MinL,MinL)) ;
    }
    int t(1) ;
    PFMImage& Img=_I ;
    
    switch(distTYPE)
    {
	        
	    case PAT:
	    {
		t= PatchSize;
		cout << " Preparing feature space for Patch metric with window half-size " << t << "... " << endl ;
		DataMat dummy(_h,_w) ;
		int nd(3*pow((2*t)+1,2)) ;
		for (int i(0); i< nd;i++) Features.push_back(dummy) ;
//		cout << Features.size() << endl ;
		for(int i(0); i<_h; i++)
		{
		    for(int j(0); j<_w; j++)
		    {
			
// 			int k1=max(i-t, 0), k2=min(i+t, _h-1) ;
// 			int l1=max(j-t, 0), l2=min(j+t, _w-1) ;
			int k1(i-t), k2(i+t) ;
			int l1(j-t), l2(j+t) ;
			int fidx(0) ;
			for(int k(k1); k<=k2; k++)
			    for(int l(l1); l<=l2; l++)
			    {
				if (k>=0 && k<_h && l>=0 && l<_w)
				{				    
				    Features[fidx](i,j) = Img(k,l).r ;
				    Features[fidx+1](i,j) = Img(k,l).g ;
				    Features[fidx+2](i,j) = Img(k,l).b ;
				}
				else
				{				 
				    Features[fidx](i,j) =  -2;
				    Features[fidx+1](i,j) = -2 ;
				    Features[fidx+2](i,j) = -2;
				}				    
				fidx+=3; 
			    }
			
		    }
		}    
	    }
	    break;

	case FAB :
	    {
	    cout << " Preparing feature space for Fabio's metric ... " << endl ;
	    
	    DataMat D1(_h, _w), D2(_h, _w), D3(_h, _w) ;
	    DataMat Dm1(_h, _w), Dm2(_h, _w), Dm3(_h, _w) ;
	    DataMat Dv1(_h, _w), Dv2(_h, _w), Dv3(_h, _w) ;
	     
	    for(int i(0); i<_h; i++)
	    {
		for(int j(0); j<_w; j++)
		{
		    D1(i,j) = _I(i,j).r ;
		    D2(i,j) = _I(i,j).g ;
		    D3(i,j) = _I(i,j).b ;
		    
		    int k1=max(i-t, 0), k2=min(i+t, _h-1) ;
		    int l1=max(j-t, 0), l2=min(j+t, _w-1) ;
		    PFMPixel MeanWin(0,0,0)  ;
		    for(int k(k1); k<=k2; k++)
			for(int l(l1); l<=l2; l++)
			    MeanWin += _I(k,l) ; 
		    MeanWin = 1.0/pow(2*t+1,2)*MeanWin ;
		    PFMPixel VarWin(0,0,0) ;
		    for(int k(k1); k<=k2; k++)
			for(int l(l1); l<=l2; l++)
			    VarWin += (_I(k,l) - MeanWin) * (_I(k,l) - MeanWin) ; 
		    VarWin = 1.0/pow(2*t+1,2)*VarWin;
		    Dm1(i,j) = MeanWin.r ;
		    Dm2(i,j) = MeanWin.g ;
		    Dm3(i,j) = MeanWin.b ;
		    Dv1(i,j) = sqrt(VarWin.r) ;
		    Dv2(i,j) = sqrt(VarWin.g) ;
		    Dv3(i,j) = sqrt(VarWin.b) ;
		}
	    }
	    
	    Features.push_back(D1) ;
	    Features.push_back(D2) ;
	    Features.push_back(D3) ;
	    Features.push_back(Dm1) ;
	    Features.push_back(Dm2) ;
	    Features.push_back(Dm3) ;
	    Features.push_back(Dv1) ;
	    Features.push_back(Dv2) ;
	    Features.push_back(Dv3) ;
	    }
	    break;
	    
	case EMD:
	case CHISQ:
/*	case EMD2:
	case CHISQ2:*/
	{	    
		t= PatchSize;
// 		t=2;		
		int nd(3*pow((2*t)+1,2)) ;
		int nBins(max(nd/10,2)) ;		
// 		int nBins(nd/9) ;
		cout << " Preparing feature space for ChiSq metric with " << nBins << " bins and patch size " << 2*t+1 << " ... " << endl ;
		
		_hists.resize(_h) ;
		_RGBhists.resize(_h) ;				
		for (int i(0); i<_h; i++) 
		{
		    _hists[i].resize(_w) ;
		    _RGBhists[i].resize(_w) ;

		    for (int j(0); j<_w; j++) 
		    {			
			_hists[i][j].Init(nBins, 0,1,3) ;
			_RGBhists[i][j].Init(nBins, 0,1) ;
		    }
		}
// 		cout << " Preparing feature space for ChiSq metric with " << nBins << " bins and patch size " << 2*t+1 << " ... " << endl ;
		#pragma omp parallel for
		for(int i=0; i<_h; i++)
		{
		    for(int j(0); j<_w; j++)
		    {
			
			int k1(i-t), k2(i+t) ;
			int l1(j-t), l2(j+t) ;
			int fidx(0) ;
			for(int k(k1); k<=k2; k++)
			    for(int l(l1); l<=l2; l++)
			    {
				if (k>=0 && k<_h && l>=0 && l<_w)
				{
				    vector<double> rgb (3,0) ;
				    rgb[0]=Img(k,l).r; rgb[1]= Img(k,l).g; rgb[2]= Img(k,l).b ;
				    _hists[i][j].Add(rgb);
				    _RGBhists[i][j].Add(.33*rgb[0]) ;
				    _RGBhists[i][j].Add(.33+rgb[1]*.33) ;
				    _RGBhists[i][j].Add(.66+.33*rgb[2]) ;
				}
			    }
			    
		    }
		}
	}
	break; 
	
    }
    _nd = Features.size() ;
    
}

DistHelper::~DistHelper() 
{
    if (ExistFIs) 
    {
	cout << "  Deleting factored image within distance helper (for scribbles) " << endl ;
	delete _pFI ;	
    }
}

double DistHelper::getDistanceSqNoScribs(int i, int j, int i2, int j2) const
{ 
    double dist(0.0) ;
    PFMPixel p = _I(i,j) - _I(i2,j2) ;
    PFMPixel p2 = _I(i,j) + _I(i2,j2) ; 
    if (fabs(p2.r)<1e-3) p2.r=1; if (fabs(p2.g)<1e-3) p2.g=1;  if (fabs(p2.b)<1e-3) p2.b=1;  
     
//     cout << p.r << " " << p.g << " " << p.b  << " " ;
    double diff = p.r*p.r+ p.g*p.g + p.b*p.b  ;
    double d2(0.0) ;
    
    switch(_dtype)
    {
/*	case CHISQ2:
	    dist = ChiSq(_hists[i][j], _hists[i2][j2]) ;
	    break; 
	    
	case EMD2:	    
		dist = EMDdist(_hists[i][j], _hists[i2][j2]) ;
		break; 
		*/
	case CHISQ:
	    dist = ChiSqOld(_hists[i][j], _hists[i2][j2]) ; 
	    break;
	
	case EMD:	    
	    dist = EMDdistOld(_hists[i][j], _hists[i2][j2]) ;
	    break ;
	    
	case L1: 
	    dist = fabs(p.r) + fabs(p.g) + fabs(p.b) ;
	    break ;

	case EUC:
	    dist = diff ;
	    break ;
	
	case BIL:
	    dist =  diff/_sigsqr + (pow(i-i2,2)+pow(j-j2,2))/_sigsqs  ;
	    break;
	    
	case FAB:	
	    for(int d(0) ; d<_nd; d++)
		d2 += pow( Features[d](i, j) - Features[d](i2,j2), 2) ; 
		dist  = d2/_sigsqr + (pow(i-i2,2)+pow(j-j2,2))/_sigsqs ;
	    break;
	    
	case PAT:
	    for(int d(0) ; d<_nd; d++)
	    {
		float f1(Features[d](i,j)), f2(Features[d](i2,j2)); 
// 		if (f1>-.1 && f2>-.1) d2 += fabs(f1-f2) ;   // Patch L1
		if (f1>-1 && f2>-1) d2 += pow(f1-f2,2.0f) ;   // Patch L2
	    }
	    dist=d2; 
	    break;
	    
	case LAP:
	    break;
	
    }
    return dist; 
}

double DistHelper::getDistanceSq(int i, int j, int i2, int j2) const
{
    double dist(0) ;
    if (ExistFIs)
    {
	double d12 (sqrt(_pFI->SqDistBetweenPixels(i,j,i2,j2))) ;
// 	double d12 (getDistanceSqNoScribs(i,j,i2,j2)) ; 
	double d1fsc (Dfs(i,j)), d1bsc(Dbs(i,j)) ;
	double d2fsc (Dfs(i2,j2)), d2bsc(Dbs(i2,j2)) ;
	
// 	dist = d12 + (d1fsc+d2fsc) + (d1bsc+d2bsc) ;
	dist = min(d12, min(d1fsc+d2fsc, d1bsc+d2bsc) ) ;
// 	dist = min(d1fsc+d2fsc, d1bsc+d2bsc)  ; 
	dist *= dist ;
    }
    else
    {
	dist = getDistanceSqNoScribs(i,j,i2,j2) ; 
    }
    return dist ;
    
}

double DistHelper::getDistance(int i, int j, int i2, int j2) const
{
    const double ret(getDistanceSq(i,j,i2,j2) ) ;
    return sqrt(ret);
}



