/*
 * Accurate binary image selection from inaccurate user input
 *
 * Authors: Kartic Subr, Sylvain Paris, Cyril Soler, Jan Kautz
 *
 * Copyright (2013):Kartic Subr. 
 * All rights reserved.
 *
 * For details, see the paper:
 * Accurate binary image selection from inaccurate user input
 * Computer Graphics Forum 32, 2 (2013)
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for educational, research, and non-commercial
 * purposes, without fee and without a signed licensing agreement, is
 * hereby granted, provided that the above copyright notice and this
 * paragraph appear in all copies modifications, and distributions.
 *
 * Any commercial use or any redistribution of this software
 * requires a license.
 *
 * For further details, contact Kartic Subr (kartic@gmail.com)
 */

#ifndef _DATAMAT_H_
#define _DATAMAT_H_

#include <vector>
#include<iostream>
#include <iterator>
#include <map>
#include <cmath>
#include <PFMImage.h>
#include <algorithm>
#include <functional>  
#include <Magick++.h> 
using namespace std; 

typedef vector<double> DataVec ;

////////////////////////////////////////////////
// Internal matrix data structure.
// An nD image embedding is stored 
// as a vector<DataMat> of length n
////////////////////////////////////////////////
class DataMat
{
	private:
		vector<vector<double> > _V ;
        
	public:                

		typedef vector<double>::iterator               iterator;
		typedef vector<double>::const_iterator         const_iterator;

		DataMat(int nRows=0, int nCols=0) ;
		DataMat(const DataMat& DM) ;
		DataMat(const DataMat& DM, int nRows, int nCols) ; // Resample DM in a nRows*nCols DataMat.
		DataMat(const string& Fname)  {Open(Fname); }
                  
                double 	MaxVal() const ;
		void		Open(const string& Fname) ;
		int             NCols() const                                   { return (_V.size()>0) ? _V[0].size() : 0 ; }
		int             y_size() const                                   { return _V.size(); }
		int             x_size() const                                   { return (_V.size()>0) ? _V[0].size() : 0 ; }
		int             NRows() const                                   { return _V.size(); }
		void            resize(int NR, int NC)                          {_V.resize(NR); for (int i(0); i<NR; i++) _V[i].resize(NC); }
		void            resize(const DataMat& D) ;                
		const DataMat&  Assign (const DataMat& D, double d) ;
		const DataMat&  operator = (const DataMat& D) ;
		const DataMat&  operator += (const DataMat& D) ;
		const DataMat&  operator -= (const DataMat& D) ;
		const DataMat&  operator *= (const DataMat& D) ;
		const DataMat&  operator -= (double d) ;
		const DataMat&  operator *= (double d) ;
		double          GetDR() const ;
		double&         operator () (const pair<int, int>& p)           { return _V[p.first][p.second] ; }                
		const double&   operator () (const pair<int, int>& p) const     { return _V[p.first][p.second] ; }                
		const double&   operator()(int i, int j) const                  { return _V[i][j] ; }
		double&         operator()(int i, int j)                        { return _V[i][j] ; }
		
		void 		Histogram(vector<double>& hist, vector<double>& CumulHist, int nBins, bool normalize=true) const; 
// 		void            PopulateWithFunc(const RBFRep& F) ;
                void            NormalizedGrad(DataMat& GradNorm) const ;
                int             InflPoints(vector<pair<int, int> >& p, double Thresh) const ;
		void		BinaryWrite(const string& fname, bool Verbose=true) const ;
		inline iterator       begin() {return _V.begin()->begin();}
		inline const_iterator begin() const {return _V.begin()->begin();}  
		inline iterator       end() {return _V.back().end();}
		inline const_iterator end() const {return _V.back().end();}
                                
		friend ostream& operator <<(ostream& os, const DataMat& Mat) ;
                
	private:
		double          Interp(double i,double j) const ;
        
};

struct less_pair: public binary_function< const pair<int,int>&, const pair<int,int>& , bool> 
{  bool operator()(const pair<int,int>& p1, const pair<int,int>& p2) { return p1.first <= p2.first && p1.second < p2.second; } };

inline ostream& operator <<(ostream& os, const DataMat& Mat) 
{
	int nR(Mat.NRows()), nC(Mat.NCols()) ;        
	for (int i(0); i<nR; i++)
	{
		for (int j(0); j<nC; j++)
		{       
			os << Mat._V[i][j] << " " ;                        
		}
		os << endl ;
	}               
	return os << endl ;
}

#endif //_DATAMAT_H_
